/*========================================================================================================
   Program Name: DetectNonPriASCII
   Purpose: Detection of non-printable and non ASCII characters
   Operation Environment: UTF-8 or Before the conversion
========================================================================================================*/
%macro DetectNonPriASCII(DataFolder=, KeyVariables=USUBJID, CVPMult=1) ;
   options nofmterr ;
   libname INDATA "&DataFolder." access=readonly cvpmult=&CVPMult. ;
   ods output members=ListOfDatasets ;
   proc datasets lib=INDATA memtype=data ;
   quit ;
   data _NULL_ ;
      set ListOfDatasets ;
      call symput('DSNAME'||strip(put(_N_,best.)),NAME) ;
      call symputx('NofDS', _N_) ;
   run ;
   data _NULL_ ;
      set ListOfDatasets ;
      call execute('proc contents data=INDATA.'||NAME||' out=ListOfVariables_'||NAME||'(keep=MEMNAME NAME TYPE LENGTH VARNUM LABEL) noprint ; run ;') ;
      call execute('proc sort data=ListOfVariables_'||NAME||' ;') ;
      call execute('   where TYPE=2 ;') ;
      call execute('   by VARNUM ; run ;') ;
      call execute('data ListOfVariables_'||NAME||' ;') ;
      call execute('   set ListOfVariables_'||NAME||' ;') ;
      call execute('   by VARNUM ;') ;
      call execute('   retain MaximumLength 0;') ;
      call execute('   if MaximumLength < Length then MaximumLength = Length ;') ;
      call execute('   call symputx("MaxLengthOf'||strip(NAME)||'",MaximumLength) ; run ;') ;
   run ;
   data ListOfASCII ;
      length ASCII_List ASCII_Control $200. ;
      ASCII_List='' ;
      do i=32 to 126 ;
         ASCII_List=trim(ASCII_List)||byte(i) ;
      end ;
      do j=1 to 31, 127 ;
         ASCII_Control=trim(ASCII_Control)||byte(j) ;
      end ;
   run ;
   data _TEMP ; stop ; run ;
   %do i=1 %to &NofDS. %by 1 ;
      proc sql ;
         create table &&DSNAME&i. as 
            select *
               from INDATA.&&DSNAME&i., ListOfASCII
      quit ;                    
      data _TEMP_&&DSNAME&i. ; stop ; run ; 
      data _NULL_ ;
         set ListOfVariables_&&DSNAME&i. ;
         call execute('data _TEMP1 ;') ;
         call execute("   set &&DSNAME&i. ;") ;
         call execute("   length MEMNAME $32. NAME $32. ORGVALUE $&&&&MaxLengthOf&&DSNAME&i.. DetectedString $&&&&MaxLengthOf&&DSNAME&i.. ;") ;
         call execute('   OBS=_N_ ;') ;
         call execute('   MEMNAME="'||strip(MEMNAME)||'"; ') ;
         call execute('   NAME="'||strip(NAME)||'"; ') ;
         call execute('   ORGVALUE=strip('||NAME||'); ') ;
         call execute('   DetectedString=kcompress('||NAME||',ASCII_List) ;') ;
         call execute('   CandidateValue="" ;') ;
         call execute('   if DetectedString^="" ;') ;
         call execute("   keep &KeyVariables. MEMNAME NAME ORGVALUE DetectedString OBS CandidateValue ; run ;") ;
         call execute("data _TEMP_&&DSNAME&i. ;") ;
         call execute("   set _TEMP_&&DSNAME&i. _TEMP1 ;") ;
         call execute("run ;") ;
         call execute('proc sql ; drop table _TEMP1 ; quit ;') ;
      run ;
   %end ;
   data _NULL_ ;
      call symput('_Date_Time', compress(put(date(),yymmdds10.),'/')||'_'||compress(put(time(),tod8.),':')) ;
   run ;
   %do i=1 %to &NofDS. %by 1 ;
      %if &i.=1 %then %do ;
         ods excel file="&_CPATH.\files\DetectNonPriASCII.xlsx" 
            options(sheet_interval='table'
                    sheet_name="&&DSNAME&i."
                    suppress_bylines='yes'
                    embedded_titles='no' ) ;
      %end ;
      %else %do ;
         ods excel 
            options(sheet_interval='table'
                    sheet_name="&&DSNAME&i."
                    suppress_bylines='yes'
                    embedded_titles='no' ) ;
      %end ;
      ods listing close;
      proc print width=min data=_TEMP_&&DSNAME&i. ;
         var &KeyVariables. OBS MEMNAME NAME ORGVALUE DetectedString CandidateValue ;
      run ;
      ods listing;
   %end ;
   ods excel close ;
   proc datasets lib=WORK kill memtype=data ;
   run ; quit ;
%mend DetectNonPriASCII ;

/*========================================================================================================
   Program Name: ReplaceNonAscii
   Purpose: To replace Non Ascii string with another string
   Operation Environment: After the conversion (UTF-8 environment)
========================================================================================================*/
%macro ReplaceNonAscii(InDataFolder=, OutDataFolder=, InExcelFile=, InExcelSheet=, OutEncoding=, CVPMult=1) ;
   options nofmterr ;
   libname INDATA  "&InDataFolder." cvpmult=&CVPMult. access=readonly ;
   libname OUTDATA "&OutDataFolder." ;
   libname EXIN excel "&InExcelFile." 
      header=yes
      dbmax_text=2000
      mixed=yes ;
   data _EXIN ;
      length CandidateValue $2000. ;
      set EXIN."&InExcelSheet.$"n ;
   run ;
   libname EXIN clear ;
   proc sort data=_EXIN       out=_EX_DATA_VAR(keep=MEMNAME OBS1 NAME CandidateValue) ;
      by MEMNAME ;
   run ; 
   data _CON_VARNUM ; stop ; run ;
   data _WK_CON ; stop ; run ;
   data _NULL_ ;
      set _EX_DATA_VAR ;
      by MEMNAME ;
      if first.MEMNAME then do ;         
         call execute('proc contents data=INDATA.'||MEMNAME||' out=_CON_VARNUM_'||MEMNAME||' noprint ; run ;') ;
         call execute('proc sort data=_CON_VARNUM_'||MEMNAME||' ;') ;
         call execute('   by VARNUM ; run ;') ;
         call execute('data _CON_VARNUM ; set _CON_VARNUM _CON_VARNUM_'||MEMNAME||' ; run ;') ;
         call execute("data _WK_"||MEMNAME||' ;') ;
         call execute('   set INDATA.'||MEMNAME||' ;') ;
      end ;
      call execute('if _N_='||OBS1||' then '||NAME||'="'||strip(CandidateValue)||'" ;') ;
      if last.MEMNAME then do ;
         call execute('run ;') ;
         call execute('data _WK_CON1_'||MEMNAME||'(keep=VARNAME VARLENGTH) ;') ;
         call execute('   set _WK_'||MEMNAME||' ;') ;
         call execute('   array CHARS _character_ ;') ;
         call execute('   length VARNAME $32. ;') ;
         call execute('   do over CHARS ;') ;
         call execute('      VARNAME=vname(CHARS) ;') ;
         call execute('      VARLENGTH=lengthn(CHARS) ;') ;
         call execute('      output ;') ;
         call execute('   end ; run ;') ;         
         call execute('proc sql ; create table _WK_CON_'||MEMNAME||' as') ;      
         call execute('   select VARNAME, max(VARLENGTH) as MAX_LENGTH, "'||strip(MEMNAME)||'" as MEMNAME') ;     
         call execute('      from _WK_CON1_'||MEMNAME||' group by VARNAME ; quit ;') ;
         call execute('data _WK_CON ; set _WK_CON _WK_CON_'||MEMNAME||' ; run ;') ; 
      end ;
   run ;   
   proc sort data=_WK_CON ; by MEMNAME VARNAME ; run ;
   data _NULL_ ;
      set _WK_CON ;
      by MEMNAME VARNAME ;
      MAX_LENGTH = ifn(MAX_LENGTH=0, 1, MAX_LENGTH) ;
      if      first.MEMNAME then do ;
         call execute('data _WK_'||MEMNAME||' ;') ;
         call execute('   length '||VARNAME||' $'||put(round(MAX_LENGTH,1.0),best.)||'. ') ;
      end ;
      else if last.MEMNAME then
         call execute(VARNAME||' $'||put(round(MAX_LENGTH,1.0),best.)||'. ; set _WK_'||MEMNAME||' ; run ;') ;
      else 
         call execute(VARNAME||' $'||put(round(MAX_LENGTH,1.0),best.)||'. ') ;
   run ;
   proc sort data=_CON_VARNUM ; by MEMNAME VARNUM ; run ;
   data _NULL_ ;
      set _CON_VARNUM ;
      by MEMNAME VARNUM ;
      if      first.MEMNAME then do ;
         call execute("data OUTDATA."||MEMNAME||"(encoding=&OutEncoding.) ;") ;
         call execute('   format '||NAME) ;
      end ;
      else if last.MEMNAME then 
         call execute(NAME||" ; set _WK_"||MEMNAME||' ; run ;') ;
      else 
         call execute(NAME) ;
   run ;
   proc datasets lib=WORK kill memtype=data ;
   run ; quit ;
%mend ReplaceNonAscii ;

/*========================================================================================================
   Program Name: Transcoding2UTF8 
   Purpose: To translate each encoding to UTF-8
   Operation Environment: UTF-8 Environment
========================================================================================================*/
%macro Transcoding2UTF8(InDataFolder=, OutDataFolder=, OutEncoding=) ;
   options nofmterr ;
   libname INDATA "&InDataFolder." cvpmultiplier=4 ;
   libname OUTDATA "&OutDataFolder." ;
   ods listing close ;
   ods output Variables=_CON_VARNUM ;
   proc contents data=INDATA._ALL_ ; run ;
   ods listing ;
   data _CON_VARNUM ;
      set _CON_VARNUM ;
      length MEMNAME NAME $32. ;
      MEMNAME=compress(tranwrd(Member,'INDATA.','')) ;
      NAME   =strip(Variable) ;
      VARNUM =NUM ;
      keep MEMNAME NAME VARNUM ;
   proc sort data=_CON_VARNUM ; by MEMNAME VARNUM ; run ;
   data _WK_CON ; stop ; run ;
   data _NULL_ ;
      set _CON_VARNUM ;
      by MEMNAME VARNUM ;
      if first.MEMNAME then do ;
         call execute('data _WK_CON1_'||MEMNAME||'(keep=VARNAME VARLENGTH) ;') ;
         call execute('   set INDATA.'||MEMNAME||' ;') ;
         call execute('   array CHARS _character_ ;') ;
         call execute('   length VARNAME $32. ;') ;
         call execute('   do over CHARS ;') ;
         call execute('      VARNAME=vname(CHARS) ;') ;
         call execute('      VARLENGTH=lengthn(CHARS) ;') ;
         call execute('      output ;') ;
         call execute('   end ; run ;') ;         
         call execute('proc sql ; create table _WK_CON_'||MEMNAME||' as') ;      
         call execute('   select VARNAME, max(VARLENGTH) as MAX_LENGTH, "'||strip(MEMNAME)||'" as MEMNAME') ;     
         call execute('      from _WK_CON1_'||MEMNAME||' group by VARNAME ; quit ;') ;
         call execute('data _WK_CON ; set _WK_CON _WK_CON_'||MEMNAME||' ; run ;') ; 
      end ;
   run ;
   proc sort data=_WK_CON ; by MEMNAME VARNAME ; run ;
   data _NULL_ ;
      set _WK_CON ;
      by MEMNAME VARNAME ;
      MAX_LENGTH = ifn(MAX_LENGTH=0, 1, MAX_LENGTH) ;
      if      first.MEMNAME then do ;
         call execute('data _WK_'||MEMNAME||' ;') ;
         call execute('   length '||VARNAME||' $'||put(round(MAX_LENGTH,1.0),best.)||'. ') ;
      end ;
      else if last.MEMNAME then
         call execute(VARNAME||' $'||put(round(MAX_LENGTH,1.0),best.)||'. ; set INDATA.'||MEMNAME||' ; format _all_ ; informat _all_ ; run ;') ;
      else 
         call execute(VARNAME||' $'||put(round(MAX_LENGTH,1.0),best.)||'. ') ;
   run ;
   data _NULL_ ;
      set _CON_VARNUM ;
      by MEMNAME VARNUM ;
      if      first.MEMNAME then do ;
         call execute("data OUTDATA."||MEMNAME||"(encoding=&OutEncoding.) ;") ;
         call execute('   format '||NAME) ;
      end ;
      else if last.MEMNAME then 
         call execute(NAME||" ; set _WK_"||MEMNAME||' ; run ;') ;
      else 
         call execute(NAME) ;
   run ;

   proc datasets lib=WORK kill memtype=data ;
   run ; quit ;
%mend Transcoding2UTF8 ;


%macro Processing ;
   %if       &_Processing.=1 %then %do ;
      %DetectNonPriASCII(DataFolder=%str(&_INDATA.), KeyVariables=%str(&_KeyVariables.), CVPMult=2) ;
   %end ;
   %else %if &_Processing.=2 %then %do ;
      %ReplaceNonAscii(InDataFolder=%str(&_INDATA.), OutDataFolder=%str(&_OUTDATA.), InExcelFile=%str(&_CPATH.\files\DetectNonPriASCII.xlsx), InExcelSheet=&_Datasets., OutEncoding=UTF8, CVPMult=2) ;
   %end ;
   %else %if &_Processing.=3 %then %do ;
      %Transcoding2UTF8(InDataFolder=%str(&_INDATA.), OutDataFolder=%str(&_OUTDATA.), OutEncoding=UTF8) ;
   %end ;
%mend Processing ;
%Processing ;
